@@ -128,6 +128,12 @@ class Normalize(DisplayedTransform):
 
     _docs=dict(encodes="Normalize batch", decodes="Denormalize batch")
 
+@typedispatch
+def show_batch(x:TensorRawImage, y, samples, ctxs=None, max_n=10, nrows=None, ncols=None, figsize=None, **kwargs):
+    if ctxs is None: ctxs = get_grid(min(len(samples), max_n), nrows=nrows, ncols=ncols, figsize=figsize)
+    ctxs = show_batch[object](x, y, samples, ctxs=ctxs, max_n=max_n, **kwargs)
+    return ctxs
+
 # vision.data
 # Cell
 @typedispatch
@@ -147,6 +153,62 @@ def show_results(x:TensorRawImage, y:TensorRawImage, samples, outs, ctxs=None, m
     ctxs[2::3] = [b.show(ctx=c, **kwargs) for b,c,_ in zip(outs.itemgot(0),ctxs[2::3],range(max_n))]
     return
 
+#for other tasks:
+
+@typedispatch
+def show_results(x:TensorRawImage, y, samples, outs, ctxs=None, max_n=10, nrows=None, ncols=None, figsize=None, **kwargs):
+    if ctxs is None: ctxs = get_grid(min(len(samples), max_n), nrows=nrows, ncols=ncols, figsize=figsize)
+    ctxs = show_results[object](x, y, samples, outs, ctxs=ctxs, max_n=max_n, **kwargs)
+    return ctxs
+
+@typedispatch
+def show_results(x:TensorRawImage, y:TensorCategory, samples, outs, ctxs=None, max_n=10, nrows=None, ncols=None, figsize=None, **kwargs):
+    if ctxs is None: ctxs = get_grid(min(len(samples), max_n), nrows=nrows, ncols=ncols, figsize=figsize)
+    for i in range(2):
+        ctxs = [b.show(ctx=c, **kwargs) for b,c,_ in zip(samples.itemgot(i),ctxs,range(max_n))]
+    ctxs = [r.show(ctx=c, color='green' if b==r else 'red', **kwargs)
+            for b,r,c,_ in zip(samples.itemgot(1),outs.itemgot(0),ctxs,range(max_n))]
+    return ctxs
+
+@typedispatch
+def show_results(x:TensorRawImage, y:TensorMask|TensorPoint|TensorBBox, samples, outs, ctxs=None, max_n=6,
+                 nrows=None, ncols=1, figsize=None, **kwargs):
+    if ctxs is None: ctxs = get_grid(min(len(samples), max_n), nrows=nrows, ncols=ncols, figsize=figsize, double=True,
+                                     title='Target/Prediction')
+    for i in range(2):
+        ctxs[::2] = [b.show(ctx=c, **kwargs) for b,c,_ in zip(samples.itemgot(i),ctxs[::2],range(2*max_n))]
+    for o in [samples,outs]:
+        ctxs[1::2] = [b.show(ctx=c, **kwargs) for b,c,_ in zip(o.itemgot(0),ctxs[1::2],range(2*max_n))]
+    return ctxs
+
+@typedispatch
+def plot_top_losses(x: TensorRawImage, y:TensorCategory, samples, outs, raws, losses, nrows=None, ncols=None, figsize=None, **kwargs):
+    axs = get_grid(len(samples), nrows=nrows, ncols=ncols, figsize=figsize, title='Prediction/Actual/Loss/Probability')
+    for ax,s,o,r,l in zip(axs, samples, outs, raws, losses):
+        s[0].show(ctx=ax, **kwargs)
+        ax.set_title(f'{o[0]}/{s[1]} / {l.item():.2f} / {r.max().item():.2f}')
+
+@typedispatch
+def plot_top_losses(x: TensorRawImage, y:TensorMultiCategory, samples, outs, raws, losses, nrows=None, ncols=None, figsize=None, **kwargs):
+    axs = get_grid(len(samples), nrows=nrows, ncols=ncols, figsize=figsize)
+    for i,(ax,s) in enumerate(zip(axs, samples)): s[0].show(ctx=ax, title=f'Image {i}', **kwargs)
+    rows = get_empty_df(len(samples))
+    outs = L(s[1:] + o + (TitledStr(r), TitledFloat(l.item())) for s,o,r,l in zip(samples, outs, raws, losses))
+    for i,l in enumerate(["target", "predicted", "probabilities", "loss"]):
+        rows = [b.show(ctx=r, label=l, **kwargs) for b,r in zip(outs.itemgot(i),rows)]
+    display_df(pd.DataFrame(rows))
+
+@typedispatch
+def plot_top_losses(x:TensorRawImage, y:TensorMask, samples, outs, raws, losses, nrows=None, ncols=None, figsize=None, **kwargs):
+    axes = get_grid(len(samples)*3, nrows=len(samples), ncols=3, figsize=figsize, flatten=False, title="Input | Target | Prediction")
+    if axes.ndim == 1: axes = (axes,)
+    titles = ["input", "target", "pred"]
+    for axs,s,o,l in zip(axes, samples, outs, losses):
+        imgs = (s[0], s[1], o[0])
+        for ax,im,title in zip(axs, imgs, titles):
+            if title=="pred": title += f"; loss = {l:.4f}"
+            im.show(ctx=ax, **kwargs)
+            ax.set_title(title)
 
 # Cell
 def _fig_bounds(x):
@@ -162,23 +224,30 @@ def show_raw_image(im, ax=None, figsize=None, title=None, ctx=None,verbose=False
     if hasattrs(im, ('data','cpu','permute')): # for Pytorch objects
         #pv("show_raw_image: (step in) hasattrs of Tensor",verbose)
         im = im.data.cpu()
-        im=((im/65535)*255) # Converting from 16bits to 8bits because stupid "imshow" can only show 8bits
+        if torch.max(im)>255:
+            im=im/65535.
+        else:
+            im/255.
+        #im=((im/65535)*255) # Converting from 16bits to 8bits because stupid "imshow" can only show 8bits
         #im=im*255
         if im.shape[0]<5: im=im.permute(1,2,0)
-    elif isinstance(im,RAWPYobj):
+    elif isinstance(im,RAWPYobj): # For Rawpy objects
         #pv("show_raw_image: (step in) isinstance of RAWPYobj",verbose)
-        im=im.ndarr # For Rawpy objects
-        im=((im/65535)*255)
+        if im.output_bps==16:
+            im=(im.ndarr/65535.)
+        else:
+            im=(im.ndarr/255.)
     elif not isinstance(im,np.ndarray):
         #pv("show_raw_image: (step in) isinstance of nd.array",verbose)
-        im=array(im) # for PIL objects?
+        im=array(im/255) # for PIL objects?
    # print(type(im))
     #print(im.shape)
     ax = ifnone(ax,ctx) # returns None if a graph wasn't initialized already
     if figsize is None: figsize = (_fig_bounds(im.shape[0]), _fig_bounds(im.shape[1]))
     if ax is None: _,ax = plt.subplots(figsize=figsize)
     #pv("show_raw_image: ax.imshow()",verbose)
-    ax.imshow(np.uint8(im), vmin=0,vmax=255, **kwargs) # prints the images on the figure
+    #ax.imshow(np.uint8(im), vmin=0,vmax=255, **kwargs) # prints the images on the figure
+    ax.imshow(im, **kwargs) # prints the images on the figure
     if title is not None: ax.set_title(title)
     ax.axis('off')
     return ax
